/*
 * kylin-status-manager
 *
 * Copyright (C) 2022, KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Authors: wangguixi <wangguixi@kylinos.cn>
 *
 */
#include <QGuiApplication>
#include <QDebug>
#include <QDBusConnection>
#include <QDBusInterface>
#include <QDBusReply>
#include <unistd.h>
#include <QThread>
#include "autoswitch.h"
#define SETTINGS_KYLINSTATUSMANAGER    "org.kylin-status-manager"
#define MODE_MONITER         "modemonitor" // 是否监听角度值
#define TABLET_MODE          "tabletmode"
#define AUTO_ROTATION        "autorotation" // 旋转
#define ROTATIONS            "rotations"
#define KEYBOARD_ENABLE      "keyboardenable" //键盘输入方式
#define FIRSTSEITCH          "firstswitch" //第一次切换

#define THEME_QT_SCHEMA "org.ukui.style"
#define THEME_STYLE_NAME "styleName"

#define MODE_TYPE       1
#define ROTATIONS_TYPE   0
#define KEYBOARD_TYPE     2
#define PCNORMAL true
AutoSwitch::AutoSwitch(QObject *parent) :
    QObject(parent)
{
    qInfo() << __FILE__ << __LINE__ << "【kylin-status-manager is running】";
    registDbus();
    QByteArray id(SETTINGS_KYLINSTATUSMANAGER);
    if(QGSettings::isSchemaInstalled(id)) {
        ksmSetting = new QGSettings(id);
    }else {
        qWarning() << __FILE__ << __LINE__ << "org.kylin-status-manager is not install";
        return;
    }
    QByteArray id3(THEME_QT_SCHEMA);
    if(QGSettings::isSchemaInstalled(id3))
    {
        themeSetting = new QGSettings(id3);
        connect(themeSetting, &QGSettings::changed, [this](const QString & key) {
            qInfo() << __FILE__ << __LINE__ << "【org.ukui.style 】changed key is:" << key;
            if(key == THEME_STYLE_NAME) {
                emit stylename_change_signal(themeSetting->get(THEME_STYLE_NAME).toString());
            }
        });
    }else {
        qWarning() << __FILE__ << __LINE__ << "org.ukui.style is not install";
    }

    //初始化设备监控
    devicestatus = new DevicesStatusMonitoring;
    if(devicestatus) {
        connect(devicestatus, &DevicesStatusMonitoring::keyboardEnableSignal,
                this, &AutoSwitch::processingKeyboardEnableSignal);
        connect(devicestatus, &DevicesStatusMonitoring::rotationsChangeSignal,
                this, &AutoSwitch::processingRotationsChangeSignal);
        connect(devicestatus, &DevicesStatusMonitoring::modeChangeSignal,
                this, &AutoSwitch::processingModeChangeSignal);
    }else {
        qWarning() << __FILE__ << __LINE__ << "设备状态监控对象为空";
    }
    //初始化配置
    init();
    //旋转翻折延时防抖处理
//    debounceTimer = new QTimer();
//    connect(debounceTimer, &QTimer::timeout, this, &AutoSwitch::modeDebounceTimeout);
    debounceTimer1 = new QTimer();
    connect(debounceTimer1, &QTimer::timeout, this, &AutoSwitch::rotationsDebounceTimeout);
    debounceTimer2 = new QTimer();
    connect(debounceTimer2, &QTimer::timeout, this, &AutoSwitch::keyboardDebounceTimeout);

    //该槽函数是为了在通过侧边栏切换时键盘状态不能更新的情况设置的。
    connect(this, &AutoSwitch::mode_change_signal, [this](bool tablet_mode){
        if(tablet_mode) {
            //判断是否是第一次切换，是则打开用户手册
            if(ksmSetting->get(FIRSTSEITCH).toBool()) {
                openUserGuide();
                ksmSetting->set(FIRSTSEITCH,false);
            }
            //将firstswitch设置为false
        }
        keyboardDebounce(100);
    });
    qInfo() << __FILE__ << __LINE__ << " 【init end】";
}
AutoSwitch::~AutoSwitch()
{
//    if(debounceTimer) {
//        delete debounceTimer;
//    }
    if(debounceTimer1) {
        delete debounceTimer1;
    }
    if(debounceTimer2) {
        delete debounceTimer2;
    }

    if(themeSetting) {
        delete themeSetting;
    }
    if(ksmSetting) {
        delete ksmSetting;
    }
    delete devicestatus;
}
void AutoSwitch::init()
{
    qDebug() << __FILE__ << __LINE__ << "【start of init】";
    if(ksmSetting) {
        ksmstruct.modeMonitor = ksmSetting->get(MODE_MONITER).toBool();//模式监控开关
        ksmstruct.autoRotation = ksmSetting->get(AUTO_ROTATION).toBool();//旋转监控开关
        ksmstruct.tabletMode = ksmSetting->get(TABLET_MODE).toBool();//机器模式状态
        ksmstruct.rotation = ksmSetting->get(ROTATIONS).toString();//旋转状态
        ksmstruct.keyboardEnable = ksmSetting->get(KEYBOARD_ENABLE).toBool();
        ksmstruct.orientationSensor = devicestatus->getOrientationSensor();
    }else {
        qWarning() << __FILE__ << __LINE__ << "【init fail】";
        return;
    }
    //初始化是否打开自动切换
    if(ksmstruct.modeMonitor == true) {
        //根据ukui-input-gather更新模式
        bool tablet_mode = devicestatus->getTabletModeStatus();
        if(tablet_mode == false) {//pc模式
            set_tabletmode(false, "ksm", "init");
            setKeyboardEnable(true, "ksm", "init");
#ifdef PCNORMAL
            if(ksmstruct.autoRotation == true) {
                set_rotation("normal", "ksm", "init");
            }
#else
            //判断自动旋转是否打开，打开则根据最新更新，未打开则设置为上一次记录的旋转状态
            if(ksmstruct.autoRotation == true) {
                updateRotationStatus();
            }else {
                set_rotation(ksmstruct.rotation, "ksm", "init");
            }
#endif
        }else {//pad模式
            set_tabletmode(true, "ksm", "init");
            setKeyboardEnable(false, "ksm", "init");
            if(ksmstruct.autoRotation == true) {
                updateRotationStatus();
            }/*else {
                set_rotation(ksmstruct.rotation, "ksm", "init");
            }*/
        }
    }else {
        //自动切换未打开，恢复上次状态
        set_tabletmode(ksmstruct.tabletMode, "ksm", "init");
        bool keyboardstatus = devicestatus->getKeyboardEnableResult();
        setKeyboardEnable(keyboardstatus, "ksm", "init");
        if(ksmstruct.tabletMode == false) {//pc
#ifdef PCNORMAL
            if(ksmstruct.autoRotation == true) {
                set_rotation("normal", "ksm", "init");
            }
#else
            //判断自动旋转是否打开，打开则根据最新更新，未打开则设置为上一次记录的旋转状态
            if(ksmstruct.autoRotation == true) {
                updateRotationStatus();
            }else {
                set_rotation(ksmstruct.rotation, "ksm", "init");
            }
#endif
        }else {//pad
            if(ksmstruct.autoRotation == true) {
                updateRotationStatus();
            }/*else {//没有开全自动旋转是根据上一次的记录还是根据默认情况设置normal??
                set_rotation(ksmstruct.rotation, "ksm", "init");//根据默认设置
            }*/
        }
    }
    qDebug() << __FILE__ << __LINE__ << "【End of init】";
}
//注册dbus
void AutoSwitch::registDbus()
{
    qDebug() << __FILE__ << __LINE__ << "【RegistDbus】";

    //被拉起后的回调
    QDBusInterface Sessioninterface("org.gonme.SessionManager",
                                    "/org/gonme/SessionManager",
                                    "org.gonme.SessionManager",
                                    QDBusConnection::sessionBus());
    //调用session的接口，通知管理服务已经启动
    QDBusReply<QString> reply = Sessioninterface.asyncCall("startupfinished",
                                                      "kylin-status-manager",
                                                      "startupfinished");

    QDBusConnection  sessionBus = QDBusConnection::sessionBus();
    if(sessionBus.registerService("com.kylin.statusmanager.interface")) {
        qDebug() << __FILE__ << __LINE__ << "Regist Dbus Service success";
        sessionBus.registerObject("/", this, QDBusConnection :: ExportAllSlots | QDBusConnection :: ExportAllSignals);
    }else {
        qCritical() << "Register DBus Service Error:" << sessionBus.lastError().message();
    }
    qDebug() << __FILE__ << __LINE__ << "【End of RegistDbus】";
}
// 处理系列
void AutoSwitch::processingModeChangeSignal(bool tabletmode)
{
    qDebug() << __FILE__ << __LINE__ << "Mode will change to :" << tabletmode;
    updateModeStatus(tabletmode);
}
void AutoSwitch::processingKeyboardEnableSignal(bool enable)
{
    m_keyboardstatus = enable;
    qDebug() << __FILE__ << __LINE__ << "Keyboard will change to :" << enable;
    keyboardDebounce(100);
}
void AutoSwitch::processingRotationsChangeSignal(QString rotations)
{
    qDebug() << __FILE__ << __LINE__ << "Rotation will change to" << rotations;
    if(ksmstruct.rotation == rotations) {
        return ;
    }
    rotationsDebounce(100);
}
void AutoSwitch::modeDebounceTimeout() //模式切换防抖(暂时不需要)
{
//    ++foldtime;
//    if (foldtime == 1) {
//        foldtime = 0;
//        debounceTimer->stop();
//        //得到最新的状态进行设置
//        updateModeStatus();
//    }
}
void AutoSwitch::rotationsDebounceTimeout() //旋转防抖
{
    ++rotetime;
    if(rotetime == 7) {
        rotetime = 0;
        debounceTimer1->stop();
        //得到最新的状态进行设置
        updateRotationStatus();
    }
}
void AutoSwitch::keyboardDebounceTimeout() //键盘缓冲
{
    ++keyboardtim;
    if (keyboardtim == 14) {
        keyboardtim = 0;
        debounceTimer2->stop();
        //得到最新的状态进行设置
        updateKeyboardStatus();
    }
}
void AutoSwitch::modeDebounce(int msec)
{
    Q_UNUSED(msec);
//    if(debounceTimer->isActive()) {
//        debounceTimer->stop();
//        foldtime = 0;
//        debounceTimer->start(msec);
//    }else {
//        debounceTimer->start(msec);
//    }
}
void AutoSwitch::rotationsDebounce(int msec)
{
    if(debounceTimer1->isActive()) {
        debounceTimer1->stop();
        keyboardtim = 0;
        debounceTimer1->start(msec);
    }else {
        debounceTimer1->start(msec);
    }
}
void AutoSwitch::keyboardDebounce(int msec)
{
    if(debounceTimer2->isActive()) {
        debounceTimer2->stop();
        rotetime = 0;
        debounceTimer2->start(msec);
    }else {
        debounceTimer2->start(msec);
    }
}
//update 系列函数
//函数中的逻辑建议不考虑其他判断，应由其他函数判断处理完成后调用使用

//根据当前键盘状态，更新当前键盘情况 ，不考虑判断
void AutoSwitch::updateKeyboardStatus()
{
    qDebug() << __FILE__ << __LINE__ << "updateKeyboardStatus" ;
//    bool keyboardenable = devicestatus->currentKeyboard();
    setKeyboardEnable(m_keyboardstatus, "ksm", "update");
}
//根据当前的gsetting更新缓存结构体数据
void AutoSwitch::updateKsmstruct()
{
    if(ksmSetting) {
        ksmstruct.modeMonitor = ksmSetting->get(MODE_MONITER).toBool();//模式监控开关
        ksmstruct.autoRotation = ksmSetting->get(AUTO_ROTATION).toBool();//旋转监控开关
        ksmstruct.tabletMode = ksmSetting->get(TABLET_MODE).toBool();//机器模式状态
        ksmstruct.rotation = ksmSetting->get(ROTATIONS).toString();//旋转状态
        ksmstruct.keyboardEnable = ksmSetting->get(KEYBOARD_ENABLE).toBool();
    }
}
//更新模式，并根据当前的设置，更新旋转的状态
void AutoSwitch::updateModeStatus(bool tablet_mode)
{
    qDebug() << __FILE__ << __LINE__ << "updateModeStatus arg:" << tablet_mode;
    if(!tablet_mode) {//pc模式
        set_tabletmode(false, "ksm", "update");
#ifdef PCNORMAL
        if(ksmstruct.autoRotation == true) {
            set_rotation("normal", "ksm", "init");
        }
#else
        //判断自动旋转是否打开，打开则根据最新更新，未打开则设置为上一次记录的旋转状态
        if(ksmstruct.autoRotation == true) {
            updateRotationStatus();
        }/*else {
            set_rotation(ksmstruct.rotation, "ksm", "init");
        }*/
#endif
    }else {
        //pad模式
        set_tabletmode(true, "ksm", "update");
        if(ksmstruct.autoRotation == true) {
            updateRotationStatus();
        }/*else {
            set_rotation(ksmstruct.rotation, "ksm", "init");
        }*/
    }
}
//更新旋转状态，并根据现在的设置，判断是否需要旋转
//updateRotationStatus的调用前提是自动旋转打开
void AutoSwitch::updateRotationStatus()
{
    qInfo() << __FILE__ << __LINE__ ;
#ifdef PCNORMAL
    if(ksmstruct.tabletMode) {
        QString rotationstatus = devicestatus->getRotationStatus();
        if(rotationstatus == "undefined") {
            qWarning() << "初始获取的值未定义";
            return;
        }
        set_rotation(rotationstatus, "ksm", "update");
    }else {
        set_rotation("normal", "ksm", "init");
    }
#else
    QString rotationstatus = devicestatus->getRotationStatus();
    set_rotation(rotationstatus, "ksm", "init");
#endif

}
//get 系列
int AutoSwitch::get_inputmethod_mode()//获得当前键盘状态情况
{ 
    qDebug() << "get_inputmethod_mode";
    if(ksmSetting) {
        if(ksmSetting->get(KEYBOARD_ENABLE).toBool()) {
            return 1;
        }else {
            return 0;
        }
    }else {
        return -1;
    }
    return 1;
}
QString AutoSwitch::get_current_stylename()//获取当前主体名称
{
    if(themeSetting) {
        return themeSetting->get(THEME_STYLE_NAME).toString();
    }else {
        return NULL;
    }
    return NULL;
}
bool AutoSwitch::get_current_tabletmode()//获得当前的模式
{
    if(ksmSetting) {
        return ksmSetting->get(TABLET_MODE).toBool();
    }else {
        return false;
    }
    return false;
}
QString AutoSwitch::get_current_rotation()//获得当前旋转情况
{
    if(ksmSetting) {
        return ksmSetting->get(ROTATIONS).toString();
    }else {
        return "normal";
    }
    return "normal";
}
bool AutoSwitch::get_modemonitor()//获取自动切换模式开关情况
{
    if(ksmSetting) {
        return ksmSetting->get(MODE_MONITER).toBool();
    }else {
        return false;
    }
    return false;
}
bool AutoSwitch::get_auto_rotation()//获取自动旋转开关情况
{
    if(ksmSetting) {
        return ksmSetting->get(AUTO_ROTATION).toBool();
    }else {
        return false;
    }
    return false;
}
/*设置接口
 * 修改采用gsetting的方式进行设置，gsetting文件仅仅用来记录
 * 通过函数调用的方式进行更改状态，增加函数复用
*/
bool AutoSwitch::set_rotation(QString rotation,QString who,QString why)
{
    qInfo() << __FILE__ << __LINE__ << who << "set_rotation => " << rotation << "to" << why;
    if(rotation != "normal" && rotation != "upside-down" && rotation != "left" && rotation != "right") {
        qWarning() << __FILE__ << __LINE__ << "rotation 参数不规范";
        return false;
    }
    //自动旋转没开则不能设置
    if(ksmstruct.autoRotation == false) {
        qInfo() << __FILE__ << __LINE__ << "【The auto rotation is not open】";
        return false;
    }
    if(ksmSetting) {
        if(ksmSetting->get(ROTATIONS).toString() == rotation) {
            qInfo() << __FILE__ << __LINE__ << "【The rotation status is same as that recorded】";
            return true;
        }
        ksmSetting->set(ROTATIONS, rotation);
        Q_EMIT rotations_change_signal(rotation);
        updateKsmstruct();
        return true;
    }else {
        return false;
    }
    return true;
}
bool AutoSwitch::set_tabletmode(bool tablet_mode,QString who,QString why)
{
    qInfo() << __FILE__ << __LINE__ << who << "set_tabletmode => " << tablet_mode << "to" << why;
    if(ksmSetting) {
        if(ksmSetting->get(TABLET_MODE).toBool() == tablet_mode) {
            qInfo() << __FILE__ << __LINE__ << "The tablet_mode status is same as that recorded";
            //如果两次切换相同，可能是手动切换后又插拔键盘，所以需要更新一下键盘状态
            updateKeyboardStatus();
            return true;
        }
        ksmSetting->set(TABLET_MODE, tablet_mode);
        Q_EMIT mode_change_signal(tablet_mode);
#ifdef PCNORMAL
        if(tablet_mode) {
            if(ksmstruct.autoRotation) {
                QString rotationstatus = devicestatus->getRotationStatus();
                set_rotation(rotationstatus,"ksm", "update");
            }
        }else {
            if(ksmstruct.rotation!="normal") {
                set_rotation("normal","ksm", "update");
            }
        }
#else
        if(ksmstruct.autoRotation) {
            QString rotationstatus = devicestatus->getRotationStatus();
            set_rotation(rotationstatus,"ksm", "update");
        }
#endif
        updateKsmstruct();
        return true;
    }else {
        return false;
    }
    return true;
}
bool AutoSwitch::set_modemonitor(bool modemonitor,QString who,QString why)
{
    qInfo() << __FILE__ << __LINE__ << who << "set_modemonitor => " << modemonitor << "to" << why;
    if(ksmSetting) {
        ksmSetting->set(MODE_MONITER,modemonitor);
        Q_EMIT modemonitor_change_signal(modemonitor);
        if(modemonitor) {
            bool tablet_mode = devicestatus->getTabletModeStatus();
            updateModeStatus(tablet_mode);
        }
        updateKsmstruct();
        return true;
    }else {
        return false;
    }
}
bool AutoSwitch::set_auto_rotation(bool autorotation,QString who,QString why)
{
    qInfo() << __FILE__ << __LINE__ << who << "set_auto_rotation = " << autorotation << "to" << why;
    if(ksmSetting) {
        if(ksmstruct.orientationSensor) {//判断是否支持方向传感器，支持自动旋转
            ksmSetting->set(AUTO_ROTATION,autorotation);
            Q_EMIT auto_rotation_change_signal(autorotation);
            if(autorotation) {
                //在打开自动旋转时确保能够设置成功，而不被上一次记录的gsetting所拦截
#ifdef PCNORMAL
                if(!ksmstruct.tabletMode) {
                    ksmSetting->set(ROTATIONS, "normal");
                    Q_EMIT rotations_change_signal("normal");
                }else {
                    QString rotationstatus = devicestatus->getRotationStatus();
                    ksmSetting->set(ROTATIONS, rotationstatus);
                    Q_EMIT rotations_change_signal(rotationstatus);
                }
#else
                QString rotationstatus = devicestatus->getRotationStatus();
                ksmSetting->set(ROTATIONS, rotationstatus);
                Q_EMIT rotations_change_signal(rotationstatus);
#endif
                updateKsmstruct();
                return true;
            }
            updateKsmstruct();
            return true;
        }else {
            return false;
        }
    }else {
        return false;
    }
    return true;
}
bool AutoSwitch::is_supported_autorotation()
{
    return ksmstruct.orientationSensor;
}

bool AutoSwitch::setKeyboardEnable(bool keyboardenable,QString who,QString why)
{
    qInfo() << __FILE__ << __LINE__ << who << "setKeyboardEnable = " << keyboardenable << "to" << why;
    if(ksmSetting) {
        ksmSetting->set(KEYBOARD_ENABLE, keyboardenable);
        Q_EMIT inputmethod_change_signal(keyboardenable);
        updateKsmstruct();
        return true;
    }else {
        return false;
    }
    return true;
}
bool AutoSwitch::openUserGuide()
{
    //打开用户手册的命令行
    qDebug() << Q_FUNC_INFO;

    QThread * openthread = new QThread;
    QString serviceName = "com.kylinUserGuide.hotel" + QString("%1%2").arg("_").arg(QString::number(getuid()));
    QDBusInterface *  iface = new QDBusInterface(serviceName,
                                                 "/",
                                                 "com.guide.hotel",
                                                 QDBusConnection::sessionBus(),
                                                 this);
    iface->moveToThread(openthread);
    openthread->start();
    QDBusPendingCall pcall = iface->asyncCall(QString("showGuide"), "hot-key");

    pcall.waitForFinished();
    openthread->exit();
    openthread->wait();
    iface->deleteLater();
    openthread->deleteLater();
    return true;
}

